"""
Subpackage ``metrics`` includes:
    - signal-to-noise (S/N) estimation,
    - S/N map generation,
    - STIMP map generation,
    - detection of point like sources (for pipelines),
    - fake companions injection,
    - fake disks generation and injection,
    - algorithms throughput estimation,
    - contrast curve generation,
    - receiver operating characteristic (ROC) curves generation.
"""
from .contrcurve import *
from .detection import *
from .fakecomp import *
from .roc import *
from .snr_source import *
from .fakedisk import *
from .scattered_light_disk import *
from .stim import *



#! /usr/bin/env python

"""
Module with routines allowing for the estimation of the uncertainty on the 
parameters of an imaged companion associated to residual speckle noise.
"""

__author__ = 'O. Wertz, C. A. Gomez Gonzalez, V. Christiaens'
__all__ = ['speckle_noise_uncertainty']

#import itertools as itt
from multiprocessing import cpu_count
import numpy as np
import matplotlib.pyplot as plt

from ..conf.utils_conf import pool_map, iterable #eval_func_tuple
from ..metrics import cube_inject_companions 
from .simplex_optim import firstguess_simplex
from .simplex_fmerit import get_mu_and_sigma
from .utils_negfc import cube_planet_free
from .mcmc_sampling import confidence


def speckle_noise_uncertainty(cube, p_true, angle_range, derot_angles, algo, 
                              psfn, plsc, fwhm, aperture_radius, 
                              indep_ap=False, cube_ref=None, fmerit='sum', 
                              algo_options={}, transmission=None, mu_sigma=None, 
                              wedge=None, weights=None, force_rPA=False, 
                              nproc=None, simplex_options=None, bins=None, 
                              save=False, output=None, verbose=True, 
                              full_output=True, plot=False):
    """
    Step-by-step procedure used to determine the speckle noise uncertainty 
    associated to the parameters of a companion candidate.

      __ 
     |   The steps 1 to 3 need to be performed for each angle.
     | 
     | - 1 - At the true planet radial distance and for a given angle, we 
     |       inject a fake companion in our planet-free cube.
     |   
     | - 2 - Then, using the negative fake companion method, we determine the 
     |       position and flux of the fake companion thanks to a Simplex  
     |       Nelder-Mead minimization.
     |
     | - 3 - We calculate the offset between the true values of the position 
     |       and the flux of the fake companion, and those obtained from the 
     |       minimization. The results will be dependent on the angular 
     |       position of the fake companion. 
     |__
     
    The resulting distribution of deviations is then used to infer the 
    1-sigma uncertainty on each parameter by fitting a 1d-gaussian.
     
    Parameters
    ----------
    cube: numpy array
        The original ADI cube.    
    p_true: tuple or numpy array with 3 elements
        The radial separation, position angle (from x=0 axis) and flux 
        associated to a given companion candidate for which the speckle 
        uncertainty is to be evaluated. The planet will first 
        be subtracted from the cube, then used for test injections.
    angle_range: 1d numpy array
        Range of angles (counted from x=0 axis, counter-clockwise) at which the 
        fake companions will be injected, in [0,360[.
    derot_angles: 1d numpy array
        Derotation angles for ADI. Length should match input cube.
    algo: python routine
        Routine to be used to model and subtract the stellar PSF. From an input
        cube, derotation angles, and optional arguments, it should return a 
        post-processed frame.
    psfn: 2d numpy array
        2d array with the normalized PSF template. The PSF image must be 
        centered wrt to the array. Therefore, it is recommended to run the 
        function ``metrics/normalize_psf()`` to generate a centered and 
        flux-normalized PSF template. 
    plsc : float
        Value of the plsc in arcsec/px. Only used for printing debug output when
        ``verbose=True``.
    fwhm: float
        FWHM of the PSF in pixels.
    aperture_radius: float
        Radius of the apertures used for NEGFC, in terms of FWHM.
    indep_ap: bool, opt.
        Whether to only consider independent apertures. If yes, will supersede
        the range provided in angle_range, and only consider the first and last
        values, then fit as many non-overlapping apertures as possible. 
        The empty cube will also be used with opposite derotation angles to 
        double the number of independent apertures.
    algo_options: dict, opt.
        Options for algo. To be provided as a dictionary. Can include ncomp 
        (for PCA), svd_mode, collapse, imlib, interpolation, scaling, delta_rot
    transmission: numpy array, optional
        Array with 2 columns. First column is the radial separation in pixels. 
        Second column is the off-axis transmission (between 0 and 1) at the 
        radial separation given in column 1.
    mu_sigma: tuple of 2 floats, bool or None, opt
        If set to None: not used, and falls back to original version of the 
        algorithm, using fmerit.
        If a tuple of 2 elements: should be the mean and standard deviation of 
        pixel intensities in an annulus centered on the lcoation of the 
        companion candidate, excluding the area directly adjacent to the CC.
        If set to anything else, but None/False/tuple: will compute said mean 
        and standard deviation automatically.
    force_rPA: bool, optional
        Whether to only search for optimal flux, provided (r,PA).
    fmerit: None
        Figure of merit to use, if mu_sigma is None.
    simplex_options: dict
        All the required simplex parameters, for instance {'tol':1e-08, 
        'max_iter':200}
    bins: int or None, opt
        Number of bins for histogram of parameter deviations. If None, will be
        determined automatically based on number of injected fake companions.
    full_output: bool, optional
        Whether to return more outputs.
    output: str, optional
        The name of the output file (if save is True)    
    save: bool, optional
        If True, the result are pickled.
    verbose: bool, optional
        If True, informations are displayed in the shell.
    plot: bool, optional
        Whether to plot the gaussian fit to the distributions of parameter 
        deviations (between retrieved and injected).
     
    Returns:
    --------
    sp_unc: numpy ndarray of 3 elements
        Uncertainties on the radius, position angle and flux of the companion,
        respectively, associated to residual speckle noise. Only 1 element if
        force_rPA is set to True.
    If full_output, also returns:
        mean_dev: numpy ndarray of 3 elements
            Mean deviation for each of the 3 parameters
        p_simplex: numpy ndarray n_fc x 3
            Parameters retrieved by the simplex for the injected fake 
            companions; n_fc is the number of injected  
        offset: numpy ndarray n_fc x 3
            Deviations with respect to the values used for injection of the 
            fake companions.
        chi2, nit, success: numpy ndarray of length n_fc 
            Outputs from the simplex function for the retrieval of the 
            parameters of each injected companion: chi square value, number of
            iterations and whether the simplex converged, respectively.
    """    
    
    if not nproc:   # Hyper-threading "duplicates" the cores -> cpu_count/2
        nproc = (cpu_count()/2)
    
    if verbose:
        print('')
        print('#######################################################')
        print('###            SPECKLE NOISE DETERMINATION          ###')
        print('#######################################################')
        print('')
    
    r_true, theta_true, f_true = p_true

    if indep_ap:
        angle_span = angle_range[-1]-angle_range[0]
        n_ap = int(np.deg2rad(angle_span)*r_true/fwhm)
        delta_theta = angle_span/n_ap
        angle_range = np.linspace(angle_range[0]+delta_theta/2, 
                                  angle_range[-1]+delta_theta/2, n_ap, 
                                  endpoint=False)

    elif angle_range[0]%360 == angle_range[-1]%360:
        angle_range = angle_range[:-1]
                  
    if verbose:              
        print('Number of steps: {}'.format(angle_range.shape[0]))
        print('')
    
    imlib = algo_options.get('imlib','opencv')
    interpolation = algo_options.get('interpolation','lanczos4')
    
    # FIRST SUBTRACT THE TRUE COMPANION CANDIDATE
    planet_parameter = np.array([[r_true, theta_true, f_true]])
    cube_pf = cube_planet_free(planet_parameter, cube, derot_angles, psfn, plsc, 
                               imlib=imlib, interpolation=interpolation,
                               transmission=transmission)

    # Measure mu and sigma once in the annulus (instead of each MCMC step)
    if isinstance(mu_sigma,tuple):
        if len(mu_sigma)!=2:
            raise TypeError("If a tuple, mu_sigma must have 2 elements")
    elif mu_sigma is not None:
        ncomp = algo_options.get('ncomp', None)
        annulus_width = algo_options.get('annulus_width', int(fwhm))
        if weights is not None:
            if not len(weights)==cube.shape[0]:
                raise TypeError("Weights should have same length as cube axis 0")
            norm_weights = weights/np.sum(weights)
        else:
            norm_weights=weights
        mu_sigma = get_mu_and_sigma(cube, derot_angles, ncomp, annulus_width, 
                                    aperture_radius, fwhm, r_true, theta_true, 
                                    cube_ref=cube_ref, wedge=wedge, algo=algo, 
                                    weights=norm_weights, 
                                    algo_options=algo_options)
      
    res = pool_map(nproc, _estimate_speckle_one_angle, iterable(angle_range), 
                   cube_pf, psfn, derot_angles, r_true, f_true, plsc, fwhm,
                   aperture_radius, cube_ref, fmerit, algo, algo_options, 
                   transmission, mu_sigma, weights, force_rPA, simplex_options, 
                   verbose=verbose)
    residuals = np.array(res)
    
    if indep_ap: # do opposite angles
        res = pool_map(nproc, _estimate_speckle_one_angle, 
                       iterable(angle_range), cube_pf, psfn, -derot_angles, 
                       r_true, f_true, plsc, fwhm, aperture_radius, cube_ref, 
                       fmerit, algo, algo_options, transmission, mu_sigma, 
                       weights, force_rPA, simplex_options, verbose=verbose)
        residuals2 = np.array(res)
        residuals = np.concatenate((residuals,residuals2))
        
    if verbose:  
        print("residuals (offsets): ", residuals[:,3], residuals[:,4],
              residuals[:,5])

    p_simplex = np.transpose(np.vstack((residuals[:,0], residuals[:,1],
                                        residuals[:,2])))
    offset = np.transpose(np.vstack((residuals[:,3],residuals[:,4],
                            residuals[:,5])))
    print(offset)
    chi2 = residuals[:,6]
    nit = residuals[:,7]
    success = residuals[:,8]

    if save:
        speckles = {'r_true':r_true,
                    'angle_range': angle_range,
                    'f_true':f_true,
                    'r_simplex':residuals[:,0],
                    'theta_simplex':residuals[:,1],
                    'f_simplex':residuals[:,2],
                    'offset': offset,
                    'chi2': chi2,
                    'nit': nit,
                    'success': success}
        
        if output is None:
            output = 'speckles_noise_result'

        from pickle import Pickler
        with open(output,'wb') as fileSave:
            myPickler = Pickler(fileSave)
            myPickler.dump(speckles)


    # Calculate 1 sigma of distribution of deviations
    print(offset.shape)
    if force_rPA:
        offset = offset[:,2]
        print(offset.shape)
    if bins is None:
        bins = int(offset.shape[0]/10)
    mean_dev, sp_unc = confidence(offset, cfd=68.27, bins=bins, 
                                  gaussian_fit=True, verbose=True, save=False, 
                                  output_dir='', force=True)
    if plot:
        plt.show()

    if full_output:
        return sp_unc, mean_dev, p_simplex, offset, chi2, nit, success
    else:
        return sp_unc


def _estimate_speckle_one_angle(angle, cube_pf, psfn, angs, r_true, f_true, 
                                plsc, fwhm, aperture_radius, cube_ref, fmerit, 
                                algo, algo_options, transmission, mu_sigma, 
                                weights, force_rPA, simplex_options, 
                                verbose=True):
                         
    if verbose:
        print('Process is running for angle: {:.2f}'.format(angle))

    cube_fc = cube_inject_companions(cube_pf, psfn, angs, flevel=f_true, 
                                     plsc=plsc, rad_dists=[r_true], 
                                     n_branches=1, theta=angle, 
                                     transmission=transmission, verbose=False)
    
    ncomp = algo_options.get('ncomp', None)
    annulus_width = algo_options.get('annulus_width', int(fwhm))
    
    res_simplex = firstguess_simplex((r_true,angle,f_true), cube_fc, angs, psfn, 
                                     plsc, ncomp, fwhm, annulus_width, 
                                     aperture_radius, cube_ref=cube_ref,
                                     fmerit=fmerit, algo=algo, 
                                     algo_options=algo_options,
                                     transmission=transmission, 
                                     mu_sigma=mu_sigma, weights=weights, 
                                     force_rPA=force_rPA, 
                                     options=simplex_options, 
                                     verbose=False)

    if force_rPA:
        simplex_res_f, = res_simplex.x
        simplex_res_r, simplex_res_PA = r_true, angle
    else:
        simplex_res_r, simplex_res_PA, simplex_res_f = res_simplex.x
    offset_r = simplex_res_r - r_true
    offset_PA = simplex_res_PA - angle
    offset_f = simplex_res_f - f_true       
    chi2 = res_simplex.fun
    nit = res_simplex.nit
    success = res_simplex.success    
    
    return (simplex_res_r, simplex_res_PA, simplex_res_f, offset_r, offset_PA, 
            offset_f, chi2, nit, success)


#! /usr/bin/env python

"""
Module with simplex (Nelder-Mead) optimization for defining the flux and 
position of a companion using the Negative Fake Companion.
"""


import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import minimize
from .simplex_fmerit import chisquare, get_mu_and_sigma
from ..pca import pca_annulus
from ..var import frame_center
from ..conf import time_ini, timing
from ..conf.utils_conf import sep


__author__ = 'O. Wertz, C. A. Gomez Gonzalez, V. Christiaens'
__all__ = ['firstguess']


def firstguess_from_coord(planet, center, cube, angs, PLSC, psf, fwhm,
                          annulus_width, aperture_radius, ncomp, cube_ref=None,
                          svd_mode='lapack', scaling=None, fmerit='sum',
                          imlib='opencv', interpolation='lanczos4',
                          collapse='median', algo=pca_annulus, delta_rot=1, 
                          algo_options={}, f_range=None, transmission=None, 
                          mu_sigma=None, weights=None, plot=False, 
                          verbose=True, save=False, debug=False):
    """ Determine a first guess for the flux of a companion at a given position
    in the cube by doing a simple grid search evaluating the reduced chi2.
    
    Parameters
    ----------
    planet: numpy.array
        The (x,y) position of the planet in the pca processed cube.
    center: numpy.array
        The (x,y) position of the cube center.
    cube: numpy.array
        The cube of fits images expressed as a numpy.array. 
    angs: numpy.array
        The parallactic angle fits image expressed as a numpy.array.         
    PLSC: float
        The platescale, in arcsec per pixel.
    psf: numpy.array
        The scaled psf expressed as a numpy.array. 
    fwhm : float
        The FHWM in pixels.           
    annulus_width: int, optional
        The width in pixels of the annulus on which the PCA is done.       
    aperture_radius: int, optional
        The radius of the circular aperture in terms of the FWHM.
    ncomp: int
        The number of principal components. 
    cube_ref : numpy ndarray, 3d, optional
        Reference library cube. For Reference Star Differential Imaging.
    svd_mode : {'lapack', 'randsvd', 'eigen', 'arpack'}, str optional
        Switch for different ways of computing the SVD and selected PCs.
    scaling : {'temp-mean', 'temp-standard'} or None, optional
        With None, no scaling is performed on the input data before SVD. With 
        "temp-mean" then temporal px-wise mean subtraction is done and with 
        "temp-standard" temporal mean centering plus scaling to unit variance 
        is done. 
    fmerit : {'sum', 'stddev'}, string optional
        Figure of merit to be used, if mu_sigma is set to None.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    collapse : {'median', 'mean', 'sum', 'trimmean', None}, str or None, optional
        Sets the way of collapsing the frames for producing a final image. If
        None then the cube of residuals is used when measuring the function of
        merit (instead of a single final frame).
    algo: python routine, opt {pca_annulus, pca_annular, pca, custom}
        Routine to be used to model and subtract the stellar PSF. From an input
        cube, derotation angles, and optional arguments, it should return a 
        post-processed frame.
    delta_rot: float, optional
        If algo is set to pca_annular, delta_rot is the angular threshold used
        to select frames in the PCA library (see description of pca_annular).
    algo_options: dict, opt
        Dictionary with additional parameters for the pca algorithm (e.g. tol,
        min_frames_lib, max_frames_lib). Note: arguments such as svd_mode,
        scaling imlib, interpolation or collapse can also be included in this
        dict (the latter are also kept as function arguments for compatibility
        with older versions of vip). 
    f_range: numpy.array, optional
        The range of flux tested values. If None, 20 values between 0 and 5000
        are tested.
    transmission: numpy array, optional
        Array with 2 columns. First column is the radial separation in pixels. 
        Second column is the off-axis transmission (between 0 and 1) at the 
        radial separation given in column 1.
    mu_sigma: tuple of 2 floats or None, opt
        If set to None: not used, and falls back to original version of the 
        algorithm, using fmerit. Otherwise, should be a tuple of 2 elements,
        containing the mean and standard deviation of pixel intensities in an 
        annulus centered on the location of the companion, excluding the area 
        directly adjacent to the companion.
    weights : 1d array, optional
        If provided, the negative fake companion fluxes will be scaled according
        to these weights before injection in the cube. Can reflect changes in 
        the observing conditions throughout the sequence.
    plot: boolean, optional
        If True, the figure chi2 vs. flux is displayed.
    verbose: boolean
        If True, display intermediate info in the shell.
    save: boolean, optional
        If True, the figure chi2 vs. flux is saved as .pdf if plot is also True
        
    Returns
    -------
    out : numpy.array
        The radial coordinates and the flux of the companion.
                
    """  
    xy = planet-center
    r0 = np.sqrt(xy[0]**2 + xy[1]**2)
    theta0 = np.mod(np.arctan2(xy[1], xy[0]) / np.pi*180, 360)

    if f_range is not None:    
        n = f_range.shape[0]
    else:
        n = 100
        f_range = np.linspace(0, 5000, n)
    
    chi2r = []
    if verbose:
        print('Step | flux    | chi2r')
        
    counter = 0
    for j, f_guess in enumerate(f_range):
        chi2r.append(chisquare((r0, theta0, f_guess), cube, angs, PLSC, psf,
                               fwhm, annulus_width, aperture_radius,
                               (r0, theta0), ncomp, cube_ref, svd_mode,
                               scaling, fmerit, collapse, algo, delta_rot,
                               imlib, interpolation, algo_options, transmission, 
                               mu_sigma, weights, debug))
        if chi2r[j] > chi2r[j-1]:
            counter += 1
        if counter == 4:
            break
        if verbose:
            print('{}/{}   {:.3f}   {:.3f}'.format(j+1, n, f_guess, chi2r[j]))

    chi2r = np.array(chi2r)
    f0 = f_range[chi2r.argmin()]  

    if plot:
        plt.figure(figsize=(8, 4))
        plt.title('$\chi^2_{r}$ vs flux')
        plt.xlim(f_range[0], f_range[:chi2r.shape[0]].max())
        plt.ylim(chi2r.min()*0.9, chi2r.max()*1.1)
        plt.plot(f_range[:chi2r.shape[0]], chi2r, linestyle='-', color='gray',
                 marker='.', markerfacecolor='r', markeredgecolor='r')
        plt.xlabel('flux')
        plt.ylabel(r'$\chi^2_{r}$')
        plt.grid('on')
    if save:
        plt.savefig('chi2rVSflux.pdf')
    if plot:
        plt.show()

    return r0, theta0, f0


def firstguess_simplex(p, cube, angs, psf, plsc, ncomp, fwhm, annulus_width, 
                       aperture_radius, cube_ref=None, svd_mode='lapack', 
                       scaling=None, fmerit='sum', imlib='opencv',
                       interpolation='lanczos4', collapse='median', 
                       algo=pca_annulus, delta_rot=1, algo_options={}, 
                       p_ini=None, transmission=None, mu_sigma=None, 
                       weights=None, force_rPA=False, options=None, 
                       verbose=False, **kwargs):
    """
    Determine the position of a companion using the negative fake companion 
    technique and a standard minimization algorithm (Default=Nelder-Mead) .
    
    Parameters
    ----------
    
    p : np.array
        Estimate of the candidate position.
    cube: numpy.array
        The cube of fits images expressed as a numpy.array. 
    angs: numpy.array
        The parallactic angle fits image expressed as a numpy.array. 
    psf: numpy.array
        The scaled psf expressed as a numpy.array.        
    plsc: float
        The platescale, in arcsec per pixel.
    ncomp: int or None
        The number of principal components.  
    fwhm : float
        The FHWM in pixels.   
    annulus_width: int, optional
        The width in pixels of the annulus on which the PCA is done.       
    aperture_radius: int, optional
        The radius of the circular aperture in terms of the FWHM.
    cube_ref : numpy ndarray, 3d, optional
        Reference library cube. For Reference Star Differential Imaging.
    svd_mode : {'lapack', 'randsvd', 'eigen', 'arpack'}, str optional
        Switch for different ways of computing the SVD and selected PCs.
    scaling : {'temp-mean', 'temp-standard'} or None, optional
        With None, no scaling is performed on the input data before SVD. With 
        "temp-mean" then temporal px-wise mean subtraction is done and with 
        "temp-standard" temporal mean centering plus scaling to unit variance 
        is done. 
    fmerit : {'sum', 'stddev'}, string optional
        Figure of merit to be used, if mu_sigma is set to None.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    collapse : {'median', 'mean', 'sum', 'trimmean', None}, str or None, optional
        Sets the way of collapsing the frames for producing a final image. If
        None then the cube of residuals is used when measuring the function of
        merit (instead of a single final frame).
    algo: python routine, opt {pca_annulus, pca_annular, pca, custom}
        Routine to be used to model and subtract the stellar PSF. From an input
        cube, derotation angles, and optional arguments, it should return a 
        post-processed frame.
    delta_rot: float, optional
        If algo is set to pca_annular, delta_rot is the angular threshold used
        to select frames in the PCA library (see description of pca_annular).
    algo_options: dict, opt
        Dictionary with additional parameters for the pca algorithm (e.g. tol,
        min_frames_lib, max_frames_lib). Note: arguments such as svd_mode,
        scaling imlib, interpolation or collapse can also be included in this
        dict (the latter are also kept as function arguments for compatibility
        with older versions of vip). 
    p_ini : np.array
        Position (r, theta) of the circular aperture center.
    transmission: numpy array, optional
        Array with 2 columns. First column is the radial separation in pixels. 
        Second column is the off-axis transmission (between 0 and 1) at the 
        radial separation given in column 1.
    mu_sigma: tuple of 2 floats or None, opt
        If set to None: not used, and falls back to original version of the 
        algorithm, using fmerit. Otherwise, should be a tuple of 2 elements,
        containing the mean and standard deviation of pixel intensities in an 
        annulus centered on the location of the companion, excluding the area 
        directly adjacent to the companion.
    weights : 1d array, optional
        If provided, the negative fake companion fluxes will be scaled according
        to these weights before injection in the cube. Can reflect changes in 
        the observing conditions throughout the sequence.
    force_rPA: bool, optional
        Whether to only search for optimal flux, provided (r,PA).
    options: dict, optional
        The scipy.optimize.minimize options.
    verbose : boolean, optional
        If True, additional information is printed out.
    **kwargs: optional
        Optional arguments to the scipy.optimize.minimize function
        
    Returns
    -------
    out : scipy.optimize.minimize solution object
        The solution of the minimization algorithm.
        
    """    
    if verbose:
        print('\nNelder-Mead minimization is running...')
     
    if p_ini is None:
        p_ini = p

    if force_rPA:
        p_t = (p[-1],)
        p_ini = (p[0],p[1])
    else:
        p_t = p
    solu = minimize(chisquare, p_t, args=(cube, angs, plsc, psf, fwhm,
                                          annulus_width, aperture_radius, p_ini,
                                          ncomp, cube_ref, svd_mode, scaling,
                                          fmerit, collapse, algo, delta_rot, 
                                          imlib, interpolation, algo_options, 
                                          transmission, mu_sigma, weights, 
                                          force_rPA),
                    method='Nelder-Mead', options=options, **kwargs)

    if verbose:
        print(solu)
    return solu
    

def firstguess(cube, angs, psfn, ncomp, plsc, planets_xy_coord, fwhm=4, 
               annulus_width=4, aperture_radius=1, cube_ref=None, 
               svd_mode='lapack', scaling=None, fmerit='sum', imlib='opencv',
               interpolation='lanczos4', collapse='median', algo=pca_annulus,
               delta_rot=1, p_ini=None, f_range=None, transmission=None, 
               mu_sigma=None, wedge=None, weights=None, force_rPA= False, 
               algo_options={}, simplex=True, simplex_options=None, plot=False, 
               verbose=True, save=False):
    """ Determines a first guess for the position and the flux of a planet.
        
    We process the cube without injecting any negative fake companion. 
    This leads to the visual detection of the planet(s). For each of them,
    one can estimate the (x,y) coordinates in pixel for the position of the 
    star, as well as the planet(s). 

    From the (x,y) coordinates in pixels for the star and planet(s), we can 
    estimate a preliminary guess for the position and flux for each planet
    by using the method "firstguess_from_coord". The argument "f_range" allows
    to indicate prior limits for the flux (optional, default: None). 
    This step can be reiterate to refine the preliminary guess for the flux.

    We can go a step further by using a Simplex Nelder_Mead minimization to
    estimate the first guess based on the preliminary guess.
           
    Parameters
    ----------
    cube: numpy.array
        The cube of fits images expressed as a numpy.array. 
    angs: numpy.array
        The parallactic angle fits image expressed as a numpy.array.  
    psfn: numpy.array
        The centered and normalized (flux in a 1*FWHM aperture must equal 1) 
        PSF 2d-array.
    ncomp: int
        The number of principal components.         
    plsc: float
        The platescale, in arcsec per pixel.  
    planets_xy_coord: array or list
        The list of (x,y) positions of the planets.
    fwhm : float, optional
        The FHWM in pixels.
    annulus_width: int, optional
        The width in pixels of the annulus on which the PCA is done.       
    aperture_radius: int, optional
        The radius of the circular aperture in terms of the FWHM.
    cube_ref : numpy ndarray, 3d, optional
        Reference library cube. For Reference Star Differential Imaging.
    svd_mode : {'lapack', 'randsvd', 'eigen', 'arpack'}, str optional
        Switch for different ways of computing the SVD and selected PCs.
    scaling : {'temp-mean', 'temp-standard'} or None, optional
        With None, no scaling is performed on the input data before SVD. With 
        "temp-mean" then temporal px-wise mean subtraction is done and with 
        "temp-standard" temporal mean centering plus scaling to unit variance 
        is done. 
    fmerit : {'sum', 'stddev'}, string optional
        Figure of merit to be used, if mu_sigma is set to None.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    collapse : {'median', 'mean', 'sum', 'trimmean', None}, str or None, opt
        Sets the way of collapsing the frames for producing a final image. If
        None then the cube of residuals is used when measuring the function of
        merit (instead of a single final frame).
    algo: python routine, opt {pca_annulus, pca_annular, pca, custom}
        Routine to be used to model and subtract the stellar PSF. From an input
        cube, derotation angles, and optional arguments, it should return a 
        post-processed frame.
    delta_rot: float, optional
        If algo is set to pca_annular, delta_rot is the angular threshold used
        to select frames in the PCA library (see description of pca_annular).
    p_ini: numpy.array
        Position (r, theta) of the circular aperture center.            
    f_range: numpy.array, optional
        The range of flux tested values. If None, 20 values between 0 and 5000
        are tested.
    transmission: numpy array, optional
        Array with 2 columns. First column is the radial separation in pixels. 
        Second column is the off-axis transmission (between 0 and 1) at the 
        radial separation given in column 1.
    mu_sigma: tuple of 2 floats, bool or None, opt
        If set to None: not used, and falls back to original version of the 
        algorithm, using fmerit.
        If a tuple of 2 elements: should be the mean and standard deviation of 
        pixel intensities in an annulus centered on the lcoation of the 
        companion candidate, excluding the area directly adjacent to the CC.
        If set to anything else, but None/False/tuple: will compute said mean 
        and standard deviation automatically.
    wedge: tuple, opt
        Range in theta where the mean and standard deviation are computed in an 
        annulus defined in the PCA image. If None, it will be calculated 
        automatically based on initial guess and derotation angles to avoid.
        If some disc signal is present elsewhere in the annulus, it is 
        recommended to provide wedge manually. The provided range should be 
        continuous and >0. E.g. provide (270, 370) to consider a PA range 
        between [-90,+10].
    weights : 1d array, optional
        If provided, the negative fake companion fluxes will be scaled according
        to these weights before injection in the cube. Can reflect changes in 
        the observing conditions throughout the sequence.
    force_rPA: bool, optional
        Whether to only search for optimal flux, provided (r,PA).
    algo_options: dict, opt
        Dictionary with additional parameters for the pca algorithm (e.g. tol,
        min_frames_lib, max_frames_lib). Note: arguments such as svd_mode,
        scaling imlib, interpolation or collapse can also be included in this
        dict (the latter are also kept as function arguments for compatibility
        with older versions of vip). 
    simplex: bool, optional
        If True, the Nelder-Mead minimization is performed after the flux grid
        search.
    simplex_options: dict, optional
        The scipy.optimize.minimize options.
    plot: boolean, optional
        If True, the figure chi2 vs. flux is displayed.
    verbose: bool, optional
        If True, display intermediate info in the shell.
    save: bool, optional
        If True, the figure chi2 vs. flux is saved.

    Returns
    -------
    out : The radial coordinates and the flux of the companion.

    Notes
    -----
    Polar angle is not the conventional NORTH-TO-EAST P.A.
    """
    if verbose:
        start_time = time_ini()
        
    planets_xy_coord = np.array(planets_xy_coord)
    n_planet = planets_xy_coord.shape[0]
    center_xy_coord = np.array(frame_center(cube[0]))

    r_0 = np.zeros(n_planet)
    theta_0 = np.zeros_like(r_0)
    f_0 = np.zeros_like(r_0)

    if weights is not None:
        if not len(weights)==cube.shape[0]:
            raise TypeError("Weights should have same length as cube axis 0")
        norm_weights = weights/np.sum(weights)
    else:
        norm_weights=weights
    
    for index_planet in range(n_planet):    
        if verbose:
            print('\n'+sep)
            print('             Planet {}           '.format(index_planet))
            print(sep+'\n')
            msg2 = 'Planet {}: flux estimation at the position [{},{}], '
            msg2 += 'running ...'
            print(msg2.format(index_planet, planets_xy_coord[index_planet, 0],
                              planets_xy_coord[index_planet, 1]))
        # Measure mu and sigma once in the annulus (instead of each MCMC step)
        if isinstance(mu_sigma,tuple):
            if len(mu_sigma)!=2:
                raise TypeError("If a tuple, mu_sigma must have 2 elements")
        elif mu_sigma is not None:
            xy = planets_xy_coord[index_planet]-center_xy_coord
            r0 = np.sqrt(xy[0]**2 + xy[1]**2)
            theta0 = np.mod(np.arctan2(xy[1], xy[0]) / np.pi*180, 360)
            mu_sigma = get_mu_and_sigma(cube, angs, ncomp, annulus_width, 
                                        aperture_radius, fwhm, r0, 
                                        theta0, cube_ref=cube_ref, 
                                        wedge=wedge, svd_mode=svd_mode, 
                                        scaling=scaling, algo=algo, 
                                        delta_rot=delta_rot, imlib=imlib, 
                                        interpolation=interpolation,
                                        collapse=collapse, weights=norm_weights, 
                                        algo_options=algo_options)
        
        res_init = firstguess_from_coord(planets_xy_coord[index_planet],
                                         center_xy_coord, cube, angs, plsc,
                                         psfn, fwhm, annulus_width,
                                         aperture_radius, ncomp,
                                         f_range=f_range, cube_ref=cube_ref,
                                         svd_mode=svd_mode, scaling=scaling,
                                         fmerit=fmerit, imlib=imlib,
                                         collapse=collapse, algo=algo, 
                                         delta_rot=delta_rot,
                                         interpolation=interpolation, 
                                         algo_options=algo_options, 
                                         transmission=transmission, 
                                         mu_sigma=mu_sigma, weights=weights,
                                         plot=plot, verbose=verbose, save=save)
        r_pre, theta_pre, f_pre = res_init

        if verbose:
            msg3 = 'Planet {}: preliminary guess: (r, theta, f)=({:.1f}, '
            msg3 += '{:.1f}, {:.1f})'
            print(msg3.format(index_planet,r_pre, theta_pre, f_pre))
        
        if simplex or force_rPA:
            if verbose:
                msg4 = 'Planet {}: Simplex Nelder-Mead minimization, '
                msg4 += 'running ...'
                print(msg4.format(index_planet))

            if simplex_options is None:
                simplex_options = {'xatol': 1e-6, 'fatol': 1e-6, 'maxiter': 800,
                                   'maxfev': 2000}
                                                         
            res = firstguess_simplex((r_pre, theta_pre, f_pre), cube, angs,
                                     psfn, plsc, ncomp, fwhm, annulus_width,
                                     aperture_radius, cube_ref=cube_ref, 
                                     svd_mode=svd_mode, scaling=scaling,
                                     fmerit=fmerit, imlib=imlib,
                                     interpolation=interpolation,
                                     collapse=collapse, algo=algo, 
                                     delta_rot=delta_rot, algo_options=algo_options, 
                                     p_ini=p_ini, transmission=transmission,
                                     mu_sigma=mu_sigma, weights=weights, 
                                     force_rPA=force_rPA, 
                                     options=simplex_options, verbose=False)
            if force_rPA:
                r_0[index_planet], theta_0[index_planet] = (r_pre, theta_pre)
                f_0[index_planet], = res.x
            else:
                r_0[index_planet], theta_0[index_planet], f_0[index_planet] = res.x
            if verbose:
                msg5 = 'Planet {}: Success: {}, nit: {}, nfev: {}, chi2r: {}'
                print(msg5.format(index_planet, res.success, res.nit, res.nfev,
                                  res.fun))
                print('message: {}'.format(res.message))
            
        else:
            if verbose:
                msg4bis = 'Planet {}: Simplex Nelder-Mead minimization skipped.'
                print(msg4bis.format(index_planet))            
            r_0[index_planet] = r_pre
            theta_0[index_planet] = theta_pre
            f_0[index_planet] = f_pre                               

        if verbose:            
            centy, centx = frame_center(cube[0])
            posy = r_0 * np.sin(np.deg2rad(theta_0[index_planet])) + centy
            posx = r_0 * np.cos(np.deg2rad(theta_0[index_planet])) + centx
            msg6 = 'Planet {}: simplex result: (r, theta, f)=({:.3f}, {:.3f}'
            msg6 += ', {:.3f}) at \n          (X,Y)=({:.2f}, {:.2f})'
            print(msg6.format(index_planet, r_0[index_planet],
                              theta_0[index_planet], f_0[index_planet],
                              posx[0], posy[0]))
    
    if verbose:
        print('\n', sep, '\nDONE !\n', sep)
        timing(start_time)

    return r_0, theta_0, f_0


__version__ = "1.0.0"

from . import andromeda
from . import preproc
from . import conf
from . import fits
from . import frdiff
from . import leastsq
from . import llsg
from . import medsub
from . import negfc
from . import nmf
from . import pca
from . import metrics
from . import specfit
from . import stats
from . import var
from .hci_dataset import *
from .hci_postproc import *
from .vip_ds9 import *



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module with utilities.
"""

__author__ = 'Carlos Alberto Gomez Gonzalez, Ralf Farkas'
__all__ = ['Progressbar',
           'check_array',
           'sep',
           'vip_figsize',
           'vip_figdpi']

import os
import sys
import numpy as np

import itertools as itt
from inspect import getargspec
from functools import wraps
from multiprocessing import Pool

from vip_hci import __version__

sep = '―' * 80
vip_figsize = (8, 5)
vip_figdpi = 100


def print_precision(array, precision=3):
    """ Prints an array with a given floating point precision. 3 by default.
    """
    return print(np.array2string(array.astype(float), precision=precision))


class SaveableEmpty(object):
    """
    Empty object. Used by ``Saveable`` to restore the state of an object without
    calling __init__. Similar to what pickle/copy do.
    """
    pass


class Saveable(object):
    def save(self, filename):
        """
        Save a VIP object to a npz file.


        """

        vip_object = self.__class__.__name__

        if hasattr(self, "_saved_attributes"):

            data = {}

            for a in self._saved_attributes:
                if hasattr(self, a):
                    data[a] = getattr(self, a)

                    # set marker to re-build the original datatype
                    # (for non-np types like float, string, ...)
                    if not isinstance(getattr(self, a), np.ndarray):
                        data["_item_{}".format(a)] = True

                np.savez_compressed(filename, _vip_version=__version__,
                                    _vip_object=vip_object, **data)

        else:
            raise RuntimeError("_saved_attributes not found for class {}"
                               "".format(vip_object))

    @classmethod
    def load(cls, filename):
        try:
            data = np.load(filename)
        except:
            data = np.load(filename + ".npz")

        if "_vip_object" not in data:
            raise RuntimeError("The file you specified is not a VIP object.")

        file_vip_object = data["_vip_object"].item()
        if file_vip_object != cls.__name__:
            raise RuntimeError("The object in the file is of type '{}', please "
                               "use that classes 'load()' method instead."
                               "".format(file_vip_object))

        file_vip_version = data["_vip_version"].item()
        if file_vip_version != __version__:
            print("The file was saved with VIP {}. There may be some"
                  "compatibility issues. Use with care."
                  "".format(file_vip_version))

        self = SaveableEmpty()
        self.__class__ = cls

        for k in data:
            if k.startswith("_"):
                continue

            if "_item_{}".format(k) in data:
                setattr(self, k, data[k].item())  # un-pack np array
            else:
                setattr(self, k, data[k])

        # add non-saved, but expected attributes (backwards compatibility)
        for exp_k in self._saved_attributes:
            if exp_k not in data:
                setattr(self, exp_k, None)

        return self


class Progressbar(object):
    """ Show progress bars. Supports multiple backends.

    Examples
    --------
    .. code:: python

        from vip_hci.var import Progressbar
        Progressbar.backend = "tqdm"

        from time import sleep

        for i in Progressbar(range(50)):
            sleep(0.02)

        # or:

        bar = Progressbar(total=50):
        for i in range(50):
            sleep(0.02)
            bar.update()

        # Progressbar can be disabled globally using
        Progressbar.backend = "hide"

        # or locally using the ``verbose`` keyword:
        Progressbar(iterable, verbose=False)

    Notes
    -----
    - `leave` keyword is natively supported by tqdm, support could be added to
      other backends too?

    """
    backend = "pyprind"

    def __new__(cls, iterable=None, desc=None, total=None, leave=True,
                backend=None, verbose=True):
        if backend is None:
            backend = Progressbar.backend

        if not verbose:
            backend = "hide"

        if backend == "tqdm":
            from tqdm import tqdm
            return tqdm(iterable=iterable, desc=desc, total=total, leave=leave,
                        ascii=True, ncols=80, file=sys.stdout,
                        bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt} [{elapsed"
                                   "}<{remaining}{postfix}]") # remove rate_fmt
        elif backend == "tqdm_notebook":
            from tqdm import tqdm_notebook
            return tqdm_notebook(iterable=iterable, desc=desc, total=total,
                                 leave=leave)
        elif backend == "pyprind":
            from pyprind import ProgBar, prog_bar
            ProgBar._adjust_width = lambda self: None  # keep constant width
            if iterable is None:
                return ProgBar(total, title=desc, stream=1)
            else:
                return prog_bar(iterable, title=desc, stream=1,
                                iterations=total)
        elif backend == "hide":
            return NoProgressbar(iterable=iterable)
        else:
            raise NotImplementedError("unknown backend")

    def set(b):
        Progressbar.backend = b


class NoProgressbar(object):
    """ Wraps an ``iterable`` to behave like ``Progressbar``, but without
    producing output.
    """
    def __init__(self, iterable=None):
        self.iterable = iterable

    def __iter__(self):
        return self.iterable.__iter__()

    def __next__(self):
        return self.iterable.__next__()

    def __getattr__(self, key):
        return self.iterable.key

    def update(self):
        pass


def algo_calculates_decorator(*calculated_attributes):
    """
    Decorator for HCIPostProcAlgo methods, describe what they calculate.
    
    There are three benefits from decorating a method:
    
    - if ``verbose=True``, prints a message about the calculated attributes and
      the ones which can be calculated next.
    - the attributes which *can* be calculated by this method are tracked, so
      if a user tries to access them *before* the function is called, an
      informative error message can be shown
    - the object knows which attributes to reset when ``run()`` is called a
      second time, on a different dataset.

    Parameters
    ----------
    *calculated_attributes : list of strings
        Strings denominating the attributes the decorated function calculates.
    
    Examples
    --------
    
    .. code:: python

        from .conf import algo_calculates_decorator as calculates

        class HCIMyAlgo(HCIPostPRocAlgo):
            def __init__(self, my_algo_param):
                self.store_args(locals())
            
            @calculates("final_frame", "snr_map")
            def run(dataset=None, verbose=True):
                frame, snr = my_heavy_calculation()
                
                self.final_frame = frame
                self.snr_map = snr

    """
    def decorator(fkt):
        @wraps(fkt)
        def wrapper(self, *args, **kwargs):
            # run the actual method
            res = fkt(self, *args, **kwargs)

            # get the kwargs the fkt sees. Note that this is a combination of
            # the *default* kwargs and the kwargs *passed* by the user
            a = getargspec(fkt)
            all_kwargs = dict(zip(a.args[-len(a.defaults):], a.defaults))
            all_kwargs.update(kwargs)
            
            if not hasattr(self, "_called_calculators"):
                self._called_calculators = []
            self._called_calculators.append(fkt.__name__)

            # show help message
            if all_kwargs.get("verbose", False):
                self._show_attribute_help(fkt.__name__)

            return res

        # set an attribute on the wrapper so _get_calculations() can find it:
        wrapper._calculates = calculated_attributes
        return wrapper
    return decorator


def check_array(input_array, dim, msg=None):
    """ Checks the dimensionality of input. In case the check is not successful,
    a TypeError is raised.

    Parameters
    ----------
    input_array : list, tuple or np.ndarray
        Input data.
    dim : int or tuple
        Number of dimensions that ``input_array`` should have. ``dim`` can take
        one of these values: 1, 2, 3, 4, (1,2), (2,3), (3,4) or (2,3,4).
    msg : str, optional
        String to be used in the error message (``input_array`` name).

    """
    if not isinstance(input_array, (list, tuple, np.ndarray)):
        raise TypeError("`input_array` must be a list, tuple of numpy ndarray")

    if msg is None:
        msg = 'Input array'
    else:
        msg = '`' + msg + '`'

    error_msg = "`dim` must be: 1, 2, 3, 4, (1,2), (2,3), (3,4) or (2,3,4)"
    if isinstance(dim, int):
        if dim < 1 or dim > 4:
            raise ValueError(error_msg)
    elif isinstance(dim, tuple):
        if dim not in ((1,2), (2,3), (3,4), (2,3,4)):
            raise ValueError(error_msg)

    msg2 = ' must be a '
    msg3 = 'd numpy ndarray'

    if dim == 1:
        if isinstance(input_array, (list, tuple)):
            input_array = np.array(input_array)
        if not isinstance(input_array, np.ndarray):
            raise TypeError(msg + msg2 + 'list, tuple or a 1' + msg3)
        if not input_array.ndim == dim:
            raise TypeError(msg + msg2 + 'list, tuple or a 1' + msg3)

    elif dim in (2, 3, 4):
        if not isinstance(input_array, np.ndarray):
            raise TypeError(msg + msg2 + str(dim) + msg3)
        else:
            if not input_array.ndim == dim:
                raise TypeError(msg + msg2 + str(dim) + msg3)

    elif isinstance(dim, tuple):
        if dim == (1, 2):
            msg_tup = '1 or 2'
        elif dim == (2, 3):
            msg_tup = '2 or 3'
        elif dim == (3, 4):
            msg_tup = '3 or 4'
        elif dim == (2, 3, 4):
            msg_tup = '2, 3 or 4'

        if isinstance(input_array, np.ndarray):
            if input_array.ndim not in dim:
                raise TypeError(msg + msg2 + msg_tup + msg3)
        else:
            raise TypeError(msg + msg2 + msg_tup + msg3)


def frame_or_shape(data):
    """
    Sanitize ``data``, always return a 2d frame.

    If ``data`` is a 2d frame, it is returned unchanged. If it is a shaped,
    return an empty array of that shape.

    Parameters
    ----------
    data : 2d ndarray or shape tuple

    Returns
    -------
    array : 2d ndarray

    """
    if isinstance(data, np.ndarray):
        array = data
        if array.ndim != 2:
            raise TypeError('`data` is not a frame or 2d array')
    elif isinstance(data, tuple):
        array = np.zeros(data)
    else:
        raise TypeError('`data` must be a tuple (shape) or a 2d array')

    return array


def eval_func_tuple(f_args):
    """ Takes a tuple of a function and args, evaluates and returns result"""
    return f_args[0](*f_args[1:])


class FixedObj(object):
    def __init__(self, v):
        self.v = v


def iterable(v):
    """ Helper function for ``pool_map``: prevents the argument from being
    wrapped in ``itertools.repeat()``.

    Examples
    --------
    .. code-block:: python

        # we have a worker function whic processes a word:

        def worker(word, method):
            # ...

        # we want to process these words in parallel fasion:
        words = ["lorem", "ipsum", "esse", "ea", "eiusmod"]

        # but all with
        method = 1

        # we then would use
        pool_map(3, worker, iterable(words), method)

        # this results in calling
        #
        # worker(words[0], 1)
        # worker(words[1], 1)
        # worker(words[2], 1)
        # ...
    """
    return FixedObj(v)


def pool_map(nproc, fkt, *args, **kwargs):
    """
    Abstraction layer for multiprocessing. When ``nproc=1``, the builtin
    ``map()`` is used. For ``nproc>1`` a ``multiprocessing.Pool`` is created.

    Parameters
    ----------
    nproc : int
        Number of processes to use.
    fkt : callable
        The function to be called with each ``*args``
    *args : function arguments
        Arguments passed to ``fkt`` By default, ``itertools.repeat`` is applied
        on all the arguments, except when you wrap the argument in
        ``iterable()``.
    msg : str or None, optional
        Description to be displayed.
    progressbar_single : bool, optional
        Display a progress bar when single-processing is used. Defaults to
        ``False``.
    verbose : bool, optional
        Show more output. Also disables the progress bar when set to ``False``.

    Returns
    -------
    res : list
        A list with the results.

    """
    msg = kwargs.get("msg", None)
    verbose = kwargs.get("verbose", True)
    progressbar_single = kwargs.get("progressbar_single", False)
    _generator = kwargs.get("_generator", False)  # not exposed in docstring

    args_r = [a.v if isinstance(a, FixedObj) else itt.repeat(a) for a in args]
    z = zip(itt.repeat(fkt), *args_r)

    if nproc == 1:
        if progressbar_single:
            total = len([a.v for a in args if isinstance(a, FixedObj)][0])
            z = Progressbar(z, desc=msg, verbose=verbose, total=total)
        res = map(eval_func_tuple, z)
        if not _generator:
            res = list(res)
    else:
        if verbose and msg is not None:
            print("{} with {} processes".format(msg, nproc))
        pool = Pool(processes=nproc)
        if _generator:
            res = pool.imap(eval_func_tuple, z)
        else:
            res = pool.map(eval_func_tuple, z)
        pool.close()
        pool.join()

    return res


def pool_imap(nproc, fkt, *args, **kwargs):
    """
    Generator version of ``pool_map``. Useful when showing a progress bar for
    multiprocessing (see examples).

    Parameters
    ----------
    nproc : int
        Number of processes to use.
    fkt : callable
        The function to be called with each ``*args``
    *args : function arguments
        Arguments passed to ``fkt``
    msg : str or None, optional
        Description to be displayed.
    progressbar_single : bool, optional
        Display a progress bar when single-processing is used. Defaults to
        ``True``.
    verbose : bool, optional
        Show more output. Also disables the progress bar when set to ``False``.

    Examples
    --------

    .. code-block:: python

        # using pool_map

        res = pool_map(2, my_worker_function, *args)

        # using pool_imap with a progessbar:

        res = list(Progressbar(pool_imap(2, my_worker_function, *args)))

    """
    kwargs["_generator"] = True
    return pool_map(nproc, fkt, *args, **kwargs)


def repeat(*args):
    """
    Applies ``itertools.repeat`` to every ``args``.


    Examples
    --------

    # instead of using

    import itertools as itt

    my_fkt(itt.repeat(a), itt.repeat(b), itt.repeat(c), d, itt.repeat(e))

    # you could use `repeat`:

    my_fkt(*repeat(a, b, c), d, *repeat(e))

    """
    return [itt.repeat(a) for a in args]


def make_chunks(l, n):
    """
    Chunks a list into ``n`` parts. The order of ``l`` is not kept. Useful for
    parallel processing when a single call is too fast, so the overhead from
    managing the processes is heavier than the calculation itself.

    Parameters
    ----------
    l : list
        Input list.
    n : int
        Number of parts.

    Examples
    --------
    .. code-block:: python

        make_chunks(range(13), 3)
            # -> [[0, 3, 6, 9, 12], [1, 4, 7, 10], [2, 5, 8, 11]]
    """
    return [l[i::n] for i in range(n)]


class redirect_output(object):
    """ Context manager for redirecting stdout/err to files"""
    def __init__(self, stdout='', stderr=''):
        self.stdout = stdout
        self.stderr = stderr

    def __enter__(self):
        self.sys_stdout = sys.stdout
        self.sys_stderr = sys.stderr

        if self.stdout:
            sys.stdout = open(self.stdout, 'w')
        if self.stderr:
            if self.stderr == self.stdout:
                sys.stderr = sys.stdout
            else:
                sys.stderr = open(self.stderr, 'w')

    def __exit__(self, exc_type, exc_value, traceback):
        # TODO: close self.stdout and self.stderr
        sys.stdout = self.sys_stdout
        sys.stderr = self.sys_stderr


def lines_of_code():
    """ Calculates the lines of code for VIP. Not objective measure of
    developer's work! (note to self).
    """
    cur_path = os.path.dirname(os.path.abspath(__file__))
    path = cur_path[:-len('conf')]

    ignore_set = set(["__init__.py"])

    loclist = []

    for pydir, _, pyfiles in os.walk(path):
        if 'exlib/' not in pydir:
            for pyfile in pyfiles:
                if pyfile not in ignore_set and pyfile.endswith(".py"):
                    totalpath = os.path.join(pydir, pyfile)
                    loclist.append(
                        (len(open(totalpath, "r").read().splitlines()),
                         totalpath.split(path)[1]))

    for linenumbercount, filename in loclist:
        print("{:05d} lines in {}".format(linenumbercount, filename))

    msg = "\nTotal: {} lines in ({}) excluding external libraries."
    print(msg.format(sum([x[0] for x in loclist]), path))



#! /usr/bin/env python
"""
Phase_function class definition
"""

__author__ = 'Julien Milli'
__all__ = []

import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import interp1d


class Phase_function(object):
    """ This class represents the scattering phase function (spf).
    It contains the attribute phase_function_calc that implements either a 
    single Henyey Greenstein phase function, a double Heyney Greenstein, 
    or any custom function (data interpolated from 
    an input list of phi, spf(phi)).
    """
    
    def __init__(self, spf_dico={'name': 'HG', 'g': 0., 'polar': False}):
        """
        Constructor of the Phase_function class. It checks whether the spf_dico
        contains a correct name and sets the attribute phase_function_calc
    
        Parameters
        ----------
        spf_dico :  dictionnary
            Parameters describing the scattering phase function to be
            implemented. By default, an isotropic phase function is implemented.
            It should at least contain the key "name" chosen between 'HG'
            (single Henyey Greenstein), 'DoubleHG' (double Heyney Greenstein) or
            'interpolated' (custom function). 
            The parameter "polar" enables to switch on the polarisation (if set 
            to True, the default is False). In this case it assumes a Rayleigh
            polarised fraction (1-(cos phi)^2) / (1+(cos phi)^2).
        """
        if not isinstance(spf_dico,dict):
            msg = 'The parameters describing the phase function must be a ' \
                  'Python dictionnary'
            raise TypeError(msg)
        if 'name' not in spf_dico.keys():
            msg = 'The dictionnary describing the phase function must contain' \
                  ' the key "name"'
            raise TypeError(msg)
        self.type = spf_dico['name']
        if 'polar' not in spf_dico.keys():
            self.polar = False
        elif not isinstance(spf_dico['polar'], bool):
            msg = 'The dictionnary describing the polarisation must be a ' \
                  'boolean'
            raise TypeError(msg)
        else: 
            self.polar = spf_dico['polar']
        if self.type == 'HG':
            self.phase_function_calc = HenyeyGreenstein_SPF(spf_dico)
        elif self.type == 'DoubleHG':
            self.phase_function_calc = DoubleHenyeyGreenstein_SPF(spf_dico)
        elif self.type == 'interpolated':
            self.phase_function_calc = Interpolated_SPF(spf_dico)
        else:
            msg = 'Type of phase function not understood: {0:s}'
            raise TypeError(msg.format(self.type))
                
    def compute_phase_function_from_cosphi(self, cos_phi):
        """
        Compute the phase function at (a) specific scattering scattering
        angle(s) phi. The argument is not phi but cos(phi) for optimization
        reasons.

        Parameters
        ----------
        cos_phi : float or array
            cosine of the scattering angle(s) at which the scattering function
            must be calculated.
        """
        phf = self.phase_function_calc.compute_phase_function_from_cosphi(cos_phi)
        if self.polar:
            return (1-cos_phi**2)/(1+cos_phi**2) * phf
        else:
            return phf

    def print_info(self):
        """
        Prints information on the type and parameters of the scattering phase 
        function.
        """
        print('----------------------------')
        print('Phase function parameters')
        print('----------------------------')
        print('Type of phase function: {0:s}'.format(self.type))
        print('Linear polarisation: {0!r}'.format(self.polar))
        self.phase_function_calc.print_info()        

    def plot_phase_function(self):
        """
        Plots the scattering phase function
        """
        phi = np.arange(0, 180, 1)
        phase_func = self.compute_phase_function_from_cosphi(np.cos(np.deg2rad(phi)))
        if self.polar:
            phase_func = (1-np.cos(np.deg2rad(phi))**2) / \
                         (1+np.cos(np.deg2rad(phi))**2) * phase_func

        plt.close(0)
        plt.figure(0)
        plt.plot(phi, phase_func)
        plt.xlabel('Scattering phase angle in degrees')
        plt.ylabel('Scattering phase function')
        plt.grid()
        plt.xlim(0, 180)
        plt.show()


class HenyeyGreenstein_SPF(object):
    """
    Implementation of a scattering phase function with a single Henyey
    Greenstein function.
    """

    def __init__(self, spf_dico={'g':0.}):
        """
        Constructor of a Heyney Greenstein phase function.
    
        Parameters
        ----------
        spf_dico :  dictionnary containing the key "g" (float)
            g is the Heyney Greenstein coefficient and should be between -1 
            (backward scattering) and 1 (forward scattering). 
        """
        # it must contain the key "g"
        if 'g' not in spf_dico.keys():
            raise TypeError('The dictionnary describing a Heyney Greenstein '
                            'phase function must contain the key "g"')
        # the value of "g" must be a float or a list of floats
        elif not isinstance(spf_dico['g'], (float, int)):
            raise TypeError('The key "g" of a Heyney Greenstein phase function'
                            ' dictionnary must be a float or an integer')
        self.set_phase_function(spf_dico['g'])

    def set_phase_function(self, g):
        """ 
        Set the value of g
        """
        if g >= 1:
            print('Warning the Henyey Greenstein parameter is greater than or '
                  'equal to 1')
            print('The value was changed from {0:6.2f} to 0.99'.format(g))
            g = 0.99
        elif g <= -1:
            print('Warning the Henyey Greenstein parameter is smaller than or '
                  'equal to -1')
            print('The value was changed from {0:6.2f} to -0.99'.format(g))
            g = -0.99
        self.g = float(g)
        
    def compute_phase_function_from_cosphi(self, cos_phi):
        """
        Compute the phase function at (a) specific scattering scattering
        angle(s) phi. The argument is not phi but cos(phi) for optimization
        reasons.

        Parameters
        ----------
        cos_phi : float or array
            cosine of the scattering angle(s) at which the scattering function
            must be calculated.
        """
        return 1./(4*np.pi)*(1-self.g**2)/(1+self.g**2-2*self.g*cos_phi)**(3./2.)         

    def print_info(self):
        """
        Prints the value of the HG coefficient
        """
        print('Heynyey Greenstein coefficient: {0:.2f}'.format(self.g))


class DoubleHenyeyGreenstein_SPF(object):
    """
    Implementation of a scattering phase function with a double Henyey
    Greenstein function.
    """
    
    def __init__(self, spf_dico={'g': [0.5,-0.3], 'weight': 0.7}):
        """
        """        
        # it must contain the key "g"
        if 'g' not in spf_dico.keys():
            raise TypeError('The dictionnary describing a Heyney Greenstein'
                            ' phase function must contain the key "g"')
        # the value of "g" must be a list of floats
        elif not isinstance(spf_dico['g'],(list,tuple,np.ndarray)):
            raise TypeError('The key "g" of a Heyney Greenstein phase '
                            'function dictionnary must be  a list of floats')
        # it must contain the key "weight"
        if 'weight' not in spf_dico.keys():
                raise TypeError('The dictionnary describing a multiple Henyey '
                                'Greenstein phase function must contain the '
                                'key "weight"')
        # the value of "weight" must be a list of floats
        elif not isinstance(spf_dico['weight'], (float, int)):
            raise TypeError('The key "weight" of a Heyney Greenstein phase '
                            'function dictionnary must be a float (weight of '
                            'the first HG coefficient between 0 and 1)')
        elif spf_dico['weight']<0 or spf_dico['weight']>1:
            raise ValueError('The key "weight" of a Heyney Greenstein phase'
                             ' function dictionnary must be between 0 and 1. It'
                             ' corresponds to the weight of the first HG '
                             'coefficient')
        if len(spf_dico['g']) != 2:
            raise TypeError('The keys "weight" and "g" must contain the same'
                            ' number of elements')
        self.g = spf_dico['g']
        self.weight = spf_dico['weight']   

    def print_info(self):
        """
        Prints the value of the HG coefficients and weights
        """
        print('Heynyey Greenstein first component : coeff {0:.2f} , '
              'weight {1:.1f}%'.format(self.g[0], self.weight*100))
        print('Heynyey Greenstein second component: coeff {0:.2f} , '
              'weight {1:.1f}%'.format(self.g[1], (1-self.weight)*100.))

    def compute_singleHG_from_cosphi(self, g, cos_phi):
        """
        Compute a single Heyney Greenstein phase function at (a) specific
        scattering scattering angle(s) phi. The argument is not phi but cos(phi)
        for optimization reasons.

        Parameters
        ----------
        g : float
            Heyney Greenstein coefficient
        cos_phi : float or array
            cosine of the scattering angle(s) at which the scattering function
            must be calculated.
        """
        return 1./(4*np.pi)*(1-g**2)/(1+g**2-2*g*cos_phi)**(3./2.)         

    def compute_phase_function_from_cosphi(self,cos_phi):
        """
        Compute the phase function at (a) specific scattering scattering
        angle(s) phi. The argument is not phi but cos(phi) for optimization
        reasons.

        Parameters
        ----------
        cos_phi : float or array
            cosine of the scattering angle(s) at which the scattering function
            must be calculated.
        """
        return self.weight * self.compute_singleHG_from_cosphi(self.g[0],
                                                               cos_phi) + \
                (1-self.weight) * self.compute_singleHG_from_cosphi(self.g[1],
                                                                    cos_phi)


class Interpolated_SPF(object):
    """
    Custom implementation of a scattering phase function by providing a list of 
    scattering phase angles and corresponding values of the phase function. 
    """
    
    def __init__(self, spf_dico={'phi':np.array([  0,  18,  36,  54,  72,  90,
                                                   108, 126, 144, 162]),
                                 'spf':np.array([3.580, 0.703, 0.141, 0.0489,
                                                 0.0233, 0.0136, 0.0091, 0.0069,
                                                 0.0056,0.005])}):
        """
        Constructor of the Interpolated_SPF class. It checks whether the spf_dico
        contains the keys 'phi' and 'spf'
    
        Parameters
        ----------
        spf_dico :  dict
            dictionnary containing at least the keys "phi" (list of scattering angles)
            and "spf" (list of corresponding scattering phase function values)
            Optionnaly it can specify the kind of interpolation requested (as 
            specified by the scipy.interpolate.interp1d function), by default
            it uses a quadratic interpolation.
        """
        for key in ['phi','spf']:
            if key not in spf_dico.keys():
                raise TypeError('The dictionnary describing a '
                                '"interpolated" phase function must contain '
                                'the key "{0:s}"'.format(key))
            elif not isinstance(spf_dico[key],(list,tuple,np.ndarray)):
                raise TypeError('The key "{0:s}" of a "interpolated" phase'
                                ' function dictionnary must be a list, np array'
                                ' or tuple'.format(key))
        if len(spf_dico['phi']) != len(spf_dico['spf']):
            raise TypeError('The keys "phi" and "spf" must contain the same'
                            ' number of elements')
        self.interpolate_phase_function(spf_dico)

    def print_info(self):
        """
        Prints the information of the spf
        """
        phi = np.linspace(0, 180, 19)
        spf = self.compute_phase_function_from_cosphi(np.cos(np.deg2rad(phi)))
        print('Scattering angle: ', phi)
        print('Interpolated scattering phase function: ', spf)

    def interpolate_phase_function(self, spf_dico):
        """
        Creates the function that returns the scattering phase function based
        on the scattering angle by interpolating the values given in the
        dictionnary. By default it uses a quadractic interpolation. 
    
        Parameters
        ----------
        spf_dico :  dict
            dictionnary containing at least the keys "phi" (list of scattering angles)
            and "spf" (list of corresponding scattering phase function values)
            Optionnaly it can specify the kind of interpolation requested (as 
            specified by the scipy.interpolate.interp1d function), by default
            it uses a quadratic interpolation.
        """
        if 'kind' in spf_dico.keys():
            if not isinstance(spf_dico['kind'],int) and spf_dico['kind'] not in \
                ['linear', 'nearest', 'zero', 'slinear', \
                 'quadratic', 'cubic', 'previous', 'next']:
                raise TypeError('The key "{0:s}" must be an integer '
                                'or a string ("linear", "nearest", "zero", "slinear", '
                                '"quadratic", "cubic", "previous",'
                                '"next"'.format(spf_dico['kind']))
            else:
                kind=spf_dico['kind']
        else:
            kind='quadratic'
        self.interpolation_function = interp1d(np.cos(np.deg2rad(spf_dico['phi'])),\
                                               spf_dico['spf'],kind=kind,\
                                               bounds_error=False,
                                               fill_value=np.nan)

    def compute_phase_function_from_cosphi(self, cos_phi):
        """
        Compute the phase function at (a) specific scattering scattering
        angle(s) phi. The argument is not phi but cos(phi) for optimization
        reasons.

        Parameters
        ----------
        cos_phi : float or array
            cosine of the scattering angle(s) at which the scattering function
            must be calculated.
        """
        return self.interpolation_function(cos_phi)



"""
Subpackage ``specfit`` has helping functions such as:

- spectral fitting
- mcmc sampling of model parameter space
- best fit search within a template libraty
- utility functions for the spectral fit
"""

from .chi import *
from .model_resampling import *
from .mcmc_sampling_spec import *
from .template_lib import *
from .utils_spec import *


"""
Subpackage ``var`` has helping functions such as:

- image filtering,
- shapes extraction (annulus, squares subimages, circular apertures),
- 2d fitting (Gaussian, Moffat).
"""


from .filters import *
from .fit_2d import *
from .shapes import *



#! /usr/bin/env python

"""
Module with frame px resampling/rescaling functions.
"""
__author__ = 'Carlos Alberto Gomez Gonzalez, V. Christiaens, R. Farkas'
__all__ = ['frame_px_resampling',
           'cube_px_resampling',
           'cube_rescaling_wavelengths',
           'check_scal_vector',
           'find_scal_vector']

import numpy as np
import warnings
try:
    import cv2
    no_opencv = False
except ImportError:
    warnings.warn("Opencv python bindings are missing.", ImportWarning)
    no_opencv = True

from scipy.ndimage.interpolation import geometric_transform, zoom
from scipy.optimize import minimize
from ..var import frame_center, get_square
from .subsampling import cube_collapse


def cube_px_resampling(array, scale, imlib='opencv', interpolation='lanczos4',
                       verbose=True):
    """
    Resample the frames of a cube with a single scale factor.

    Wrapper of ``frame_px_resampling``. Useful when we need to upsample
    (upscaling) or downsample (pixel binning) a set of frames, e.g. an ADI cube.

    Parameters
    ----------
    array : 3d numpy ndarray
        Input cube, 3d array.
    scale : int, float or tuple
        Scale factor for upsampling or downsampling the frames in the cube. If
        a tuple it corresponds to the scale along x and y.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_px_resampling``
        function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_px_resampling``
        function.
    verbose : bool, optional
        Whether to print out additional info such as the new cube shape.

    Returns
    -------
    array_resc : numpy ndarray
        Output cube with resampled frames.

    """
    if array.ndim != 3:
        raise TypeError('Input array is not a cube or 3d array.')

    array_resc = []
    for i in range(array.shape[0]):
        imresc = frame_px_resampling(array[i], scale=scale, imlib=imlib,
                                     interpolation=interpolation)
        array_resc.append(imresc)

    array_resc = np.array(array_resc)

    if verbose:
        print("Cube successfully rescaled")
        print("New shape: {}".format(array_resc.shape))
    return array_resc


def frame_px_resampling(array, scale, imlib='opencv', interpolation='lanczos4',
                        verbose=False):
    """
    Resample the pixels of a frame wrt to the center, changing the frame size.

    If ``scale`` < 1 then the frame is downsampled and if ``scale`` > 1 then its
    pixels are upsampled.

    Parameters
    ----------
    array : numpy ndarray
        Input frame, 2d array.
    scale : int, float or tuple
        Scale factor for upsampling or downsampling the frame. If a tuple it
        corresponds to the scale along x and y.
    imlib : {'ndimage', 'opencv'}, optional
        Library used for image transformations.
    interpolation : str, optional
        For 'ndimage' library: 'nearneig', bilinear', 'biquadratic', 'bicubic',
        'biquartic', 'biquintic'. The 'nearneig' interpolation is the fastest
        and the 'biquintic' the slowest. The 'nearneig' is the worst
        option for interpolation of noisy astronomical images.
        For 'opencv' library: 'nearneig', 'bilinear', 'bicubic', 'lanczos4'.
        The 'nearneig' interpolation is the fastest and the 'lanczos4' the
        slowest and accurate.
    verbose : bool, optional
        Whether to print out additional info such as the new image shape.

    Returns
    -------
    array_resc : numpy ndarray
        Output resampled frame.

    """
    if array.ndim != 2:
        raise TypeError('Input array is not a frame or 2d array')

    if isinstance(scale, tuple):
        scale_x, scale_y = scale
    elif isinstance(scale, (float, int)):
        scale_x = scale
        scale_y = scale
    else:
        raise TypeError('`scale` must be float, int or tuple')

    if imlib == 'ndimage':
        if interpolation == 'nearneig':
            order = 0
        elif interpolation == 'bilinear':
            order = 1
        elif interpolation == 'biquadratic':
            order = 2
        elif interpolation == 'bicubic':
            order = 3
        elif interpolation == 'biquartic' or interpolation == 'lanczos4':
            order = 4
        elif interpolation == 'biquintic':
            order = 5
        else:
            raise TypeError('Scipy.ndimage interpolation method not recognized')

        array_resc = zoom(array, zoom=(scale_y, scale_x), order=order)

    elif imlib == 'opencv':
        if no_opencv:
            msg = 'Opencv python bindings cannot be imported. Install opencv or'
            msg += ' set imlib to ndimage'
            raise RuntimeError(msg)

        if interpolation == 'bilinear':
            intp = cv2.INTER_LINEAR
        elif interpolation == 'bicubic':
            intp = cv2.INTER_CUBIC
        elif interpolation == 'nearneig':
            intp = cv2.INTER_NEAREST
        elif interpolation == 'lanczos4':
            intp = cv2.INTER_LANCZOS4
        else:
            raise TypeError('Opencv interpolation method not recognized')

        array_resc = cv2.resize(array.astype(np.float32), (0, 0), fx=scale_x,
                                fy=scale_y, interpolation=intp)

    else:
        raise ValueError('Image transformation library not recognized')

    array_resc /= scale_y * scale_x

    if verbose:
        print("Image successfully rescaled")
        print("New shape: {}".format(array_resc.shape))

    return array_resc


def cube_rescaling_wavelengths(cube, scal_list, full_output=True, inverse=False,
                               y_in=None, x_in=None, imlib='opencv',
                               interpolation='lanczos4', collapse='median',
                               pad_mode='reflect'):
    """
    Scale/Descale a cube by scal_list, with padding.

    Wrapper to scale or descale a cube by factors given in scal_list,
    without any loss of information (zero-padding if scaling > 1).
    Important: in case of IFS data, the scaling factors in scal_list should be
    >= 1 (ie. provide the scaling factors as for scaling to the longest
    wavelength channel).

    Parameters
    ----------
    cube: 3D-array
       Data cube with frames to be rescaled.
    scal_list: 1D-array
       Vector of same dimension as the first dimension of datacube, containing
       the scaling factor for each frame.
    full_output: bool, optional
       Whether to output just the rescaled cube (False) or also its median,
       the new y and x shapes of the cube, and the new centers cy and cx of the
       frames (True).
    inverse: bool, optional
       Whether to inverse the scaling factors in scal_list before applying them
       or not; i.e. True is to descale the cube (typically after a first scaling
       has already been done)
    y_in, x_in: int
       Initial y and x sizes, required for ``inverse=True``. In case the cube is
       descaled, these values will be used to crop back the cubes/frames to
       their original size.
    imlib : {'opencv', 'ndimage'}, str optional
        Library used for image transformations. Opencv is faster than ndimage or
        skimage.
    interpolation : str, optional
        For 'ndimage' library: 'nearneig', bilinear', 'bicuadratic', 'bicubic',
        'biquartic', 'biquintic'. The 'nearneig' interpolation is the fastest
        and the 'biquintic' the slowest. The 'nearneig' is the poorer
        option for interpolation of noisy astronomical images.
        For 'opencv' library: 'nearneig', 'bilinear', 'bicubic', 'lanczos4'.
        The 'nearneig' interpolation is the fastest and the 'lanczos4' the
        slowest and accurate. 'lanczos4' is the default.
    collapse : {'median', 'mean', 'sum', 'trimmean'}, str optional
        Sets the way of collapsing the frames for producing a final image.
    pad_mode : str, optional
        One of the following string values:

            ``'constant'``
                pads with a constant value
            ``'edge'``
                pads with the edge values of array
            ``'linear_ramp'``
                pads with the linear ramp between end_value and the array edge
                value.
            ``'maximum'``
                pads with the maximum value of all or part of the vector along
                each axis
            ``'mean'``
                pads with the mean value of all or part of the vector along each
                axis
            ``'median'``
                pads with the median value of all or part of the vector along
                each axis
            ``'minimum'``
                pads with the minimum value of all or part of the vector along
                each axis
            ``'reflect'``
                pads with the reflection of the vector mirrored on the first and
                last values of the vector along each axis
            ``'symmetric'``
                pads with the reflection of the vector mirrored along the edge
                of the array
            ``'wrap'``
                pads with the wrap of the vector along the axis. The first
                values are used to pad the end and the end values are used to
                pad the beginning

    Returns
    -------
    frame: 2d array
        The median of the rescaled cube.
    cube : 3d array
        [full_output] rescaled cube
    frame : 2d array
        [full_output] median of the rescaled cube
    y,x,cy,cx : float
        [full_output] New y and x shapes of the cube, and the new centers cy and
        cx of the frames

    """
    n, y, x = cube.shape

    max_sc = np.amax(scal_list)

    if not inverse and max_sc > 1:
        new_y = int(np.ceil(max_sc * y))
        new_x = int(np.ceil(max_sc * x))
        if (new_y - y) % 2 != 0:
            new_y += 1
        if (new_x - x) % 2 != 0:
            new_x += 1
        pad_len_y = (new_y - y) // 2
        pad_len_x = (new_x - x) // 2
        pad_width = ((0, 0), (pad_len_y, pad_len_y), (pad_len_x, pad_len_x))
        big_cube = np.pad(cube, pad_width, pad_mode)
    else:
        big_cube = cube.copy()

    n, y, x = big_cube.shape
    cy, cx = frame_center(big_cube[0])

    if inverse:
        scal_list = 1. / scal_list
        cy, cx = frame_center(cube[0])

    # (de)scale the cube, so that a planet would now move radially
    cube = _cube_resc_wave(big_cube, scal_list, ref_xy=(cx, cy),
                           imlib=imlib, interpolation=interpolation)
    frame = cube_collapse(cube, collapse)

    if inverse and max_sc > 1:
        if y_in is None or x_in is None:
            raise ValueError("You need to provide y_in and x_in when "
                             "inverse=True!")
        siz = max(y_in, x_in)
        if frame.shape[0] > siz:
            frame = get_square(frame, siz, cy, cx)
        if full_output:
            n_z = cube.shape[0]
            array_old = cube.copy()
            cube = np.zeros([n_z, siz, siz])
            for zz in range(n_z):
                cube[zz] = get_square(array_old[zz], siz, cy, cx)

    if full_output:
        return cube, frame, y, x, cy, cx
    else:
        return frame


def _scale_func(output_coords, ref_xy=0, scaling=1.0, scale_y=None,
                scale_x=None):
    """
    For each coordinate point in a new scaled image (output_coords),
    coordinates in the image before the scaling are returned. This scaling
    function is used within geometric_transform which, for each point in the
    output image, will compute the (spline) interpolated value at the
    corresponding frame coordinates before the scaling.
    """
    ref_x, ref_y = ref_xy
    if scale_y is None:
        scale_y = scaling
    if scale_x is None:
        scale_x = scaling
    return (ref_y + (output_coords[0] - ref_y) / scale_y,
            ref_x + (output_coords[1] - ref_x) / scale_x)


def _frame_rescaling(array, ref_xy=None, scale=1.0, imlib='opencv',
                     interpolation='lanczos4', scale_y=None, scale_x=None):
    """
    Rescale a frame by a factor wrt a reference point.

    The reference point is by default the center of the frame (typically the
    exact location of the star). However, it keeps the same dimensions.

    Parameters
    ----------
    array : numpy ndarray
        Input frame, 2d array.
    ref_xy : float, optional
        Coordinates X,Y  of the point wrt which the rescaling will be
        applied. By default the rescaling is done with respect to the center
        of the frame.
    scale : float
        Scaling factor. If > 1, it will upsample the input array equally
        along y and x by this factor.
    scale_y : float
        Scaling factor only for y axis. If provided, it takes priority on
        scale parameter.
    scale_x : float
        Scaling factor only for x axis. If provided, it takes priority on
        scale parameter.

    Returns
    -------
    array_out : numpy ndarray
        Resulting frame.

    """
    if array.ndim != 2:
        raise TypeError('Input array is not a frame or 2d array.')

    if scale_y is None:
        scale_y = scale
    if scale_x is None:
        scale_x = scale

    outshape = array.shape
    if ref_xy is None:
        ref_xy = frame_center(array)

    if imlib == 'ndimage':
        if interpolation == 'nearneig':
            order = 0
        elif interpolation == 'bilinear':
            order = 1
        elif interpolation == 'biquadratic':
            order = 2
        elif interpolation == 'bicubic':
            order = 3
        elif interpolation == 'biquartic' or interpolation == 'lanczos4':
            order = 4
        elif interpolation == 'biquintic':
            order = 5
        else:
            raise TypeError(
                'Scipy.ndimage interpolation method not recognized')

        array_out = geometric_transform(array, _scale_func, order=order,
                                        output_shape=outshape,
                                        extra_keywords={'ref_xy': ref_xy,
                                                        'scaling': scale,
                                                        'scale_y': scale_y,
                                                        'scale_x': scale_x})

    elif imlib == 'opencv':
        if no_opencv:
            msg = 'Opencv python bindings cannot be imported. Install '
            msg += ' opencv or set imlib to skimage'
            raise RuntimeError(msg)

        if interpolation == 'bilinear':
            intp = cv2.INTER_LINEAR
        elif interpolation == 'bicubic':
            intp = cv2.INTER_CUBIC
        elif interpolation == 'nearneig':
            intp = cv2.INTER_NEAREST
        elif interpolation == 'lanczos4':
            intp = cv2.INTER_LANCZOS4
        else:
            raise TypeError('Opencv interpolation method not recognized')

        M = np.array([[scale_x, 0, (1. - scale_x) * ref_xy[0]],
                      [0, scale_y, (1. - scale_y) * ref_xy[1]]])
        array_out = cv2.warpAffine(array.astype(np.float32), M, outshape,
                                   flags=intp)

    else:
        raise ValueError('Image transformation library not recognized')

    array_out /= scale_y * scale_x
    return array_out


def _cube_resc_wave(array, scaling_list, ref_xy=None, imlib='opencv',
                    interpolation='lanczos4', scaling_y=None, scaling_x=None):
    """
    Rescale a cube by factors from ``scaling_list`` wrt a position.

    Parameters
    ----------
    array : numpy ndarray
        Input 3d array, cube.
    scaling_list : 1D-array
        Scale corresponding to each frame in the cube.
    ref_xy : float, optional
        Coordinates X,Y of the point with respect to which the rescaling will be
        performed. By default the rescaling is done with respect to the center
        of the frames; central pixel if the frames have odd size.
    imlib : str optional
        See the documentation of ``vip_hci.preproc.cube_rescaling_wavelengths``.
    interpolation : str, optional
        See the documentation of ``vip_hci.preproc.cube_rescaling_wavelengths``.
    scaling_y : 1D-array or list
        Scaling factor only for y axis. If provided, it takes priority on
        scaling_list.
    scaling_x : 1D-array or list
        Scaling factor only for x axis. If provided, it takes priority on
        scaling_list.

    Returns
    -------
    array_sc : numpy ndarray
        Resulting cube with rescaled frames.

    """

    if array.ndim != 3:
        raise TypeError('Input array is not a cube or 3d array')

    array_sc = []
    if scaling_list is None:
        scaling_list = [None]*array.shape[0]
    for i in range(array.shape[0]):
        array_sc.append(_frame_rescaling(array[i], ref_xy=ref_xy,
                                         scale=scaling_list[i], imlib=imlib,
                                         interpolation=interpolation,
                                         scale_y=scaling_y, scale_x=scaling_x))
    return np.array(array_sc)


def check_scal_vector(scal_vec):
    """
    Turn wavelengths (IFS data) into a scaling factor list.

    It checks that it has the right format: all scaling factors should be >= 1
    (i.e. the scaling should be done wrt the longest wavelength of the cube).

    Parameters
    ----------
    scal_vec: 1d array or list
        Vector with the wavelengths.

    Returns
    -------
    scal_vec: numpy ndarray, 1d
        Vector containing the scaling factors (after correction to comply with
        the condition >= 1).

    """
    if not isinstance(scal_vec, (list, np.ndarray)):
        raise TypeError('`Scal_vec` is neither a list or an np.ndarray')

    scal_vec = np.array(scal_vec)

    # checking if min factor is 1:
    if scal_vec.min() != 1:
        scal_vec = 1 / scal_vec
        scal_vec /= scal_vec.min()

    return scal_vec


def find_scal_vector(cube, lbdas, fluxes, mask=None, nfp=2, fm="stddev", 
                     simplex_options=None, debug=False, **kwargs):
    """
    Find the optimal scaling factor for the channels of an IFS cube (or .

    The algorithm finds the optimal scaling factor that minimizes residuals in
    the rescaled frames. It takes the inverse of the wavelength vector as a 
    first guess, and uses a similar method as the negative fake companion 
    technique, but minimizing residuals in either a mask or the whole field.

    Parameters
    ----------
    cube: 3D-array
       Data cube with frames to be rescaled.
    lbdas: 1d array or list
        Vector with the wavelengths, used for first guess on scaling factor.
    fluxes: 1d array or list
        Vector with the (unsaturated) fluxes at the different wavelengths, 
        used for first guess on flux factor.
    mask: 2D-array, opt
        Binary mask, with ones where the residual intensities should be 
        evaluated. If None is provided, the whole field is used.
    nfp: int, opt, {1,2}
        Number of free parameters: spatial scaling alone or spatial scaling + 
        flux scaling.
    fm: str, opt, {"sum","stddev"}
        Figure of merit to use: sum of squared residuals or stddev of residual 
        pixels.
    options: dict, optional
        The scipy.optimize.minimize options.
    **kwargs: optional
        Optional arguments to the scipy.optimize.minimize function
        
    Returns
    -------
    scal_vec: numpy ndarray, 1d
        Vector containing the scaling factors (after correction to comply with
        the condition >= 1).

    """

    scal_vec_ini = lbdas[-1]/lbdas
    n_z = len(lbdas)
    if n_z != len(fluxes) or n_z != cube.shape[0]:
        msg = "first axis of cube, fluxes and lbda must have same length"
        raise TypeError(msg)

    if simplex_options is None:
        simplex_options = {'xatol': 1e-6, 'fatol': 1e-6, 'maxiter': 800,
                           'maxfev': 2000}
    scal_vec = np.ones(n_z)
    flux_vec = np.ones(n_z)
    for z in range(n_z-1):
        flux_scal = fluxes[-1]/fluxes[z]
        cube_tmp = np.array([cube[z],cube[-1]])
        if nfp==1:
            p_ini = (scal_vec_ini[z],)
            solu = minimize(_chisquare_scal, p_ini, args=(cube_tmp, flux_scal, 
                                                          mask, fm),
                            method='Nelder-Mead', options=simplex_options, 
                            **kwargs)
            scal_fac, =  solu.x
            flux_fac = flux_scal
        else:
            p_ini = (scal_vec_ini[z],flux_scal)
            solu = minimize(_chisquare_scal_2fp, p_ini, args=(cube_tmp, mask, fm),
                            method='Nelder-Mead', options=simplex_options, 
                            **kwargs)      
            scal_fac, flux_fac =  solu.x
        if debug:
            print("channel {:.0f}:".format(z), solu.x)
        scal_vec[z] = scal_fac
        flux_vec[z] = flux_fac

    scal_vec = check_scal_vector(scal_vec)
    
    return scal_vec, flux_vec


def _find_indices_sdi(wl, dist, index_ref, fwhm, delta_sep=1, nframes=None,
                      debug=False):
    """
    Find optimal wavelengths which minimize self-subtraction in model PSF
    subtraction.

    Parameters
    ----------
    wl : numpy ndarray or list
        Vector with the scaling factors.
    dist : float
        Separation or distance (in pixels) from the center of the array.
    index_ref : int
        The `wl` index for which we are finding the pairs.
    fwhm : float
        Mean FWHM of all the wavelengths (in pixels).
    delta_sep : float, optional
        The threshold separation in terms of the mean FWHM.
    nframes : None or int, optional
        Must be an even value. In not None, then between 2 and adjacent
        ``nframes`` are kept.
    debug : bool, optional
        It True it prints out debug information.

    Returns
    -------
    indices : numpy ndarray
        List of good indices.

    """
    wl = np.asarray(wl)
    wl_ref = wl[index_ref]
    sep_lft = (wl_ref - wl) / wl_ref * ((dist + fwhm * delta_sep) / fwhm)
    sep_rgt = (wl - wl_ref) / wl_ref * ((dist - fwhm * delta_sep) / fwhm)
    map_lft = sep_lft >= delta_sep
    map_rgt = sep_rgt >= delta_sep
    indices = np.nonzero(map_lft | map_rgt)[0]

    if debug:
        print("dist: {}, index_ref: {}".format(dist, index_ref))
        print("sep_lft:", "  ".join(["{:+.2f}".format(x) for x in sep_lft]))
        print("sep_rgt:", "  ".join(["{:+.2f}".format(x) for x in sep_rgt]))
        print("indices:", indices)
        print("indices size: {}".format(indices.size))

    if indices.size == 0:
        raise RuntimeError("No frames left after radial motion threshold. Try "
                           "decreasing the value of `delta_sep`")

    if nframes is not None:
        i1 = map_lft.sum()
        window = nframes // 2
        if i1 - window < 0 or i1 + window > indices[-1]:
            window = nframes
        ind1 = max(0, i1 - window)
        ind2 = min(wl.size, i1 + window)
        indices = indices[ind1: ind2]

        if indices.size < 2:
            raise RuntimeError("No frames left after radial motion threshold. "
                               "Try decreasing the value of `delta_sep` or "
                               "`nframes`")

    if debug:
        print("indices (nframes):", indices)

    return indices


def _chisquare_scal(modelParameters, cube, flux_fac=1, mask=None, fm='sum'):
    """
    Calculate the reduced chi2:
    \chi^2_r = \frac{1}{N-3}\sum_{j=1}^{N} |I_j|,
    where N is the number of pixels within a circular aperture centered on the 
    first estimate of the planet position, and I_j the j-th pixel intensity.
    
    Parameters
    ----------    
    modelParameters: tuple
        The model parameters, typically (scal_fac, flux_fac).
    cube: numpy.array
        The cube of fits images expressed as a numpy.array.
    mask: 2D-array, opt
        Binary mask, with ones where the residual intensities should be 
        evaluated. If None is provided, the whole field is used.
    fm: str, opt, {"sum","stddev"}
        Figure of merit to use: sum of squared residuals or stddev of residual 
        pixels.
        
    Returns
    -------
    chi: float
        The reduced chi squared.
        
    """
    # rescale in flux and spatially
    array = cube.copy()
    #scale_fac, flux_fac = modelParameters
    scale_fac, = modelParameters
    array[0]*=flux_fac
    scaling_list = np.array([scale_fac,1])
    array = _cube_resc_wave(array, scaling_list)

    frame = array[1]-array[0]
    if mask is None:
        mask = np.ones_like(frame)
    
    
    if fm == 'sum':
        chi = np.sum(np.power(frame[np.where(mask)],2))
    elif fm == 'stddev':
        values = frame[np.where(mask)]
        values = values[values != 0]
        chi = np.std(values)
    else:
        raise RuntimeError('fm choice not recognized.')
        
    return chi

def _chisquare_scal_2fp(modelParameters, cube, mask=None, fm='sum'):
    """
    Calculate the reduced chi2:
    \chi^2_r = \frac{1}{N-3}\sum_{j=1}^{N} |I_j|,
    where N is the number of pixels within a circular aperture centered on the 
    first estimate of the planet position, and I_j the j-th pixel intensity.
    
    Parameters
    ----------    
    modelParameters: tuple
        The model parameters, typically (scal_fac, flux_fac).
    cube: numpy.array
        The cube of fits images expressed as a numpy.array.
    mask: 2D-array, opt
        Binary mask, with ones where the residual intensities should be 
        evaluated. If None is provided, the whole field is used.
    fm: str, opt, {"sum","stddev"}
        Figure of merit to use: sum of squared residuals or stddev of residual 
        pixels.
        
    Returns
    -------
    chi: float
        The reduced chi squared.
        
    """
    # rescale in flux and spatially
    array = cube.copy()
    scale_fac, flux_fac = modelParameters
    array[0]*=flux_fac
    scaling_list = np.array([scale_fac,1])
    array = _cube_resc_wave(array, scaling_list)

    frame = array[1]-array[0]
    if mask is None:
        mask = np.ones_like(frame)
    
    
    if fm == 'sum':
        chi = np.sum(np.power(frame[np.where(mask)],2))
    elif fm == 'stddev':
        values = frame[np.where(mask)]
        values = values[values != 0]
        chi = np.std(values)
    else:
        raise RuntimeError('fm choice not recognized.')
        
    return chi


"""
Robust linear fitting.
"""


__author__ = 'Ralf Farkas'
__all__ = []


import numpy as np


def fitaffine(x, y, debug=False):
    """
    Optimize affine equation ``y=bx+a`` in a robust way.

    This procedure calculates the best parameters ``a`` and ``b`` that optimize
    the affine equation ``y=bx+a`` in a robust way from a set of point
    ``(xi, yi)``.


    Parameters
    ----------
    x,y : 1d numpy ndarray
        The data to be fitted in a robust affine optimisation. Note the
        "reversed" order of the parameters, compared to the IDL implementation
        (see notes).
    debug : bool, optional
        Show debug output.

    Returns
    -------
    b,a : floats
        parameters which satisfy ``y = bx + a``.
        ``b`` corresponds to ``gamma`` in ANDROMEDA, ``a`` is ``gamma_prime``.

    Notes
    -----
    - ported and adapted from ``LibAndromeda/fitaffine.pro`` v1.1 2016/02/16
    - IDL version adapted from "Numerical Recipies (3rd, 2007)", p.818
    - the ``abdev`` return value was removed, as not used in ANDROMEDA.

    Pay attention to the order of the parameters!

    .. code:: IDL

        FITAFFINE, DATA_INPUT=[[y],[x]], GAMMA_OUTPUT=g1

    .. code:: python

        g1 = fitaffine(x, y)

    """
    ndata = x.shape[0]

    if debug:
        print("FITAFFINE: ***next dataset***")

    # first guess for a and b (LS):
    sx = np.sum(x)
    sy = np.sum(y)
    sxy = np.sum(x*y)
    sxx = np.sum(x**2)

    delta = ndata*sxx - sx**2
    a_ls = (sxx*sy - sx*sxy)/delta
    b_ls = (ndata*sxy - sx*sy)/delta

    if debug:
        print("FITAFFINE: first guess LS: {} + {} x".format(a_ls, b_ls))

    # chi-square to choose the first iteration step:
    chisq = np.sum((y - (a_ls + b_ls*x))**2)
    sigb = np.sqrt(chisq/delta)

    # guessed bracket at 3 sigma for b:
    a = a_ls
    b = b_ls
    b1 = b_ls
    f1, a = rofunc(x=x, y=y, b=b1)
    if debug:
        print("FITAFFINE: entering iteration loop")

    if sigb > 0 and f1 != 0:
        if f1 > 0:
            b2 = b1 + 3*sigb
        else:
            b2 = b1 - 3*sigb

        f2, a = rofunc(x, y, b=b2)

        # bracketing
        while f1*f2 > 0:
            b = b2 + 1.6*(b2-b1)
            b1 = b2
            f1 = f2
            b2 = b
            f2, a = rofunc(x, y, b=b2)

        # bisection:
        sigb = 0.01*sigb
        while np.abs(b2-b1) > sigb:
            b = b1 + 0.5*(b2-b1)
            f, a = rofunc(x, y, b=b)
            if f*f1 >= 0:
                f1 = f
                b1 = b
            else:
                f2 = f
                b2 = b

    if debug:
        print("FITAFFINE: *end of iterative loop*")
        print("FITAFFINE: equation of the robust fit: {} + {} x".format(a, b))

    return b, a


def rofunc(x, y, b):
    """
    Compute the affine parameter for robust affine fit.

    This function calculates the parameter ``a`` for a given value of ``b``
    that solves the equation ``0 = Sum_i (Xi * SIGN(Yi-a-b*Xi))`` where
    ``Xi`` is the points-set to be fitted by the point set ``Yi`` following
    ``Y=A+BX``. This function takes place in the framework of a robust affine
    fit.

    Parameters
    ----------
    x : numpy ndarray
        the fitted data set
        in IDL: DATA_INPUT[0]
    y : numpy ndarray
        the original data to be fitted
        in IDL: DATA_INPUT[1]
    b : float
        The known value ``b`` to calculate ``a``.

    Returns
    -------
    sum_result : float
    a : float
        The computed value of ``a``

    Notes
    -----
    - ported from LibAndromeda/rofunc.pro v1.1 2016/02/18
    - removed ``abdev`` variable, which is not useful here.
    - IDL version adapted from Numerical Recipies 3rd edition (2007) - p.818

    """
    epsilon = 1e-5  # convergence criteria

    arr = y - b*x
    a = np.median(arr)

    sum_result = 0
    for j in range(len(x)):
        d = y[j] - (b*x[j] + a)
        if y[j] != 0:
            d /= np.abs(y[j])
        if np.abs(d) > epsilon:
            if d >= 0:
                sum_result += x[j]
            else:
                sum_result -= x[j]

    return sum_result, a
